package analyzer

import "github.com/alecthomas/participle/v2"
import "github.com/alecthomas/participle/v2/lexer"
import "git.tebibyte.media/sashakoshka/fspl/entity"
import "git.tebibyte.media/sashakoshka/fspl/parser"

// Tree is a semantic tree. It contains the same constructs as the syntax tree,
// but with their semantic information filled in.
type Tree struct {
	rawTypes     map[string] *entity.Typedef
	rawFunctions map[string] *entity.Function
	rawMethods   map[string] *entity.Method
	ast parser.Tree

	Types     map[string] *entity.Type
	Functions map[string] *entity.Function
}

// Analyze takes in an AST and analyzes its contents within the context of this
// semantic tree. Analyzed entities will be added to the tree.
func (this *Tree) Analyze (ast parser.Tree) error {
	this.ensure()
	this.ast = ast
	err := this.assembleRawMaps()
	if err != nil { return err }
	return this.analyzeDeclarations()
}

func (this *Tree) assembleRawMaps () error {
	for _, declaration := range this.ast.Declarations {
		switch declaration.(type) {
		case *entity.Typedef:
			ty := declaration.(*entity.Typedef)
			err := this.rawNameUnique(ty.Pos, ty.Name)
			if err != nil { return err }
			this.rawTypes[ty.Name] = ty
		case *entity.Function:
			function := declaration.(*entity.Function)
			err := this.rawNameUnique (
				function.Pos,
				function.Signature.Name)
			if err != nil { return err }
			this.rawFunctions[function.Signature.Name] = function
		case *entity.Method:
			method := declaration.(*entity.Method)
			name := method.TypeName + "." + method.Signature.Name
			err := this.rawNameUnique(method.Pos, name)
			if err != nil { return err }
			this.rawMethods[name] = method
		}
	}
	return nil
}

func (this *Tree) rawNameUnique (currentPos lexer.Position, name string) error {
	var pos    lexer.Position
	var unique = true

	if ty, isType := this.rawTypes[name]; isType {
		pos = ty.Pos
		unique  = false
	}
	if function, isFunction := this.rawFunctions[name]; isFunction {
		pos = function.Pos
		unique  = false
	}
	if method, isMethod := this.rawMethods[name]; isMethod {
		pos = method.Pos
		unique  = false
	}

	if unique {
		return nil
	} else {
		return participle.Errorf (
			currentPos, "%s already declared at %v",
			name, pos)
	}
}

func (this *Tree) analyzeDeclarations () error {
	// TODO
	return nil
}

func (this *Tree) ensure () {
	if this.rawTypes != nil { return }
	this.rawTypes     = make(map[string] *entity.Typedef)
	this.rawFunctions = make(map[string] *entity.Function)
	this.rawMethods   = make(map[string] *entity.Method)
	this.Types     = make(map[string] *entity.Type)
	this.Functions = make(map[string] *entity.Function)
}
